#include <stdbool.h>
#include <stdint.h>

typedef const char *NativeString;

typedef unsigned long long NativeHandle;

typedef NativeString (*WalletPasswordCallback)(void);

typedef void (*ExternLogCallbackType)(unsigned int, NativeString);

typedef uint32_t *UIntRefMut;

typedef NativeString (*ByteArrayOutCallback)(UIntRefMut);

typedef NativeString (*ByteArrayInByteArrayOutCallback)(NativeString, UIntRefMut);

#ifdef __cplusplus
extern "C" {
#endif // __cplusplus

///
/// # Parameters
/// * `agent_handle`: The native handle created by `init_agent` call (MUST)
/// * `new_wallet_password_callback`: The callback that requests the new password from the user
/// # Return
/// `Ok(())` if the password has been changed successfully
///
/// # Remarks
/// This function doesn't override the previous password callback that has been set during the initialization of the wallet.
/// In fact the "old" password is requested via the inital callback that was set to this instance during initialization.
/// Therefore, you are responsible to send the "old" password when the initial callback is called during the update
///
/// After the update has been finished, the `new_wallet_password_callback` is dropped. Therefore, after the update, you are supposed to
/// send the **new password** from the `password_callback` that was provided during the agent creation
///
/// (See [`crate::externs::create_open::create_agent`])
///
NativeString change_password(NativeHandle agent_handle,
                             WalletPasswordCallback new_wallet_password_callback);

///
/// Create a did and its keypair in the wallet
///
/// # Parameters
/// * `agent_handle`: The native handle created by `init_agent` call (MUST)
/// * `seed`: A 32 Characters long String that will be used as a SEED for generation of the keypair (OPTIONAL)
/// # Returns
/// The newly created did as a DidInfo
///
/// # Example
///
/// ```
/// const char *local_did = create_did(agent_handle, "000000000000000MyLovelyDid000001");
/// printf("%s", local_did);
/// ```
/// __Output__
/// ```
/// {
///   "code": "OK",
///   "value": {
///     "id": "TDzjPekxmm8cmoiv6AbDWh",
///     "metadata": null,
///     "publicDID": false,
///     "public_key": "FHzfPH62yMaTBZbVoBU9aTifW1u3wcB8B5jEcB4zNGzk",
///     "seed": "4F7BsTMVPKFshM1MwLf6yJ1PgDXegnwCRCtUs4sotFsE"
///   }
/// }
/// ```
///
NativeString create_did(NativeHandle agent_handle,
                        NativeString seed);

///
/// List all the DID's existing in the wallet
///
/// # Parameters
/// * `agent_handle`: The native handle created by `init_agent` call (MUST)
///
/// # Returns
/// A list of DidInfos
///
NativeString list_dids(NativeHandle agent_handle);

/// Get the DEFAULT did (the initial did that is created during agent creation)
/// # Parameters
/// * `agent_handle`: The native handle created by `init_agent` call (MUST)
/// # Returns
/// The default did if exists, Nothing otherwise
NativeString get_default_did(NativeHandle agent_handle);

/// # sign
/// Sign the given data with my defauld `did`
///
/// ## Parameters
/// * `agent_handle`: The agent handle
/// * `data_base64`: The data to be signed
/// ## Return
/// The return value is an SDK result containing the base 64 encoded signature
///
NativeString sign(NativeHandle agent_handle, NativeString data_base64);

/// # sign_with_did
/// Sign the given data (encoded as `base64` in `data_base64`) with
/// the private key of the `did` that exists in my wallet.
///
/// # Parameters
/// * `agent_handle`: The agent handle
/// * `did`: The `DID` to sign (its private key must exist in the wallet)
/// * `data_base64`: The data to be signed
/// # Return
/// The return value is an SDK result containing the
/// base 64 encoded signature
///
///
NativeString sign_with_did(NativeHandle agent_handle, NativeString did, NativeString data_base64);

///
/// # verify_with_did_public_key
/// Given the signed data, the signature and the DID of the signer,
/// and the public key of the signer, perform a signature verification directly
///
/// # Parameters
/// * agent_handle: The native handle of the agent object
/// * did_public_key_base58: **Base 58** encoded PUBLIC KEY of the signer
/// * data_base64: Base64 encoded data that has been signed
/// * signature_base64: Base64 encoded signature
///
/// # Returns
///
/// The result of the verification (boolean value `true/false`) as and SDK Result
///
NativeString verify_with_did_public_key(NativeHandle agent_handle,
                                        NativeString did_public_key_base58,
                                        NativeString data_base64,
                                        NativeString signature_base64);

///
/// # get_did_info
/// Given the `did` return the `DIDInfo` assoicated.
///
/// ```
/// pub struct DIDInfo {
///    pub id: String,
///    pub public_key: String,
///    pub seed: Option<String>,
///    #[serde(rename = "publicDID")]
///    pub publicDID: bool,
///    pub metadata: Option<HashMap<String, String>>,
/// }
/// ```
///
/// # Parameters
/// * agent_handle: The native handle of the agent object
/// * did: The did for the DIDInfo
///
/// # Returns
/// The resolved `DIDInfo`
///
NativeString get_did_info(NativeHandle agent_handle, NativeString did);

///
/// Store the data (which is given as base64) with the given `key`.
/// # Parameters
/// * agent_handle: The native handle of the agent object
/// `key`: The identifier of the data
/// `data`: base64 encoded data
/// # Returns
/// The result of the verification (boolean value `true/false`) as and SDK Result
///
NativeString wallet_store_data(NativeHandle agent_handle, NativeString key, NativeString data);

///
/// search the wallet for the data with the given `key` and return a base64 encoded string if the data exists.
/// # Parameters
/// `agent_handle`: The handle of the agent
/// `key`: The identifier of the data
/// # Returns
///
/// If the data is found, a sample like below is returned
/// ```
/// {
///   "code": "OK",
///   "value": "base64 encoded data"
/// }
///```
///
NativeString wallet_get_data(NativeHandle agent_handle,
                             NativeString key);

///
/// Create and initialize, or open and initialize the agent
///
/// # Parameters
/// * `alias`: The alias for the agent.
/// * `seed`: Optional seed for default did creation.
/// * `password_callback`: A callback that requests password from the user in order to access
/// the sensitive data in the wallet (MUST)
///
/// If the agent is created/opened successfully and initialization result json is returned.
/// The data structure is as follows
/// ```
/// typedef struct _NativeSession {
///   NativeHandle handle,
/// } NativeSession, *PNativeSession;
/// ```
/// The handle value will then be used throughout the entire api as `agent_handle`
/// (the first parameter of almost all the functions)
///
/// # Example Output
/// ```
/// {
///   "code": "OK",
///   "value": {
///     "handle": 1453
///    }
/// }
/// ```
/// # Remarks
/// If the agent cannot be created, please check out the error code.
///
///
NativeString create_agent(NativeString alias,
                          NativeString seed,
                          WalletPasswordCallback wallet_password_callback);

///
/// initialize logging level.
///
/// 0 is off
///
/// 1 is ERR
///
/// 2 is warn
///
/// 3 is info
///
/// 4 is debug
///
/// 5 and more is trace
///
/// `callback` should be function pointer that takes an unsigned int (level) and a const char * (log message)
///
/// ## Example
///  ```
/// unsigned int level=3;
///
/// void log_message(int level, const char* log_message){
///   printf("%d %s", level, log_message);
/// }
///
/// init_logging(level, log_message);
///
/// ```
/// In platforms like Java you can bind it to a logging framework
///
NativeString init_logging(unsigned int level,
                          ExternLogCallbackType callback);

///
/// Set akche home path for all agents.
///
/// # Parameters
/// * `local_dir`: The **absolute** file path.
///
/// # Example
/// ```
/// const *char path = "/home/yerlibilgin/myAgentFilesLocation"
/// set_akche_home_path(char_path);
/// ```
///
NativeString set_akche_home_path(NativeString local_dir);

/// Explain the detail of the given code
///
/// # Parameters
/// * `error_code`: The code that returned by the API in an error case
///
/// # Return
/// The detailed description of the error
///
/// # Example
///
/// ```
/// const *char some_code = "WalletLoginFailed"
/// printf("%s", some_code, explaing_error_code(some_code));
/// ```
/// __Output__
/// ```
/// {
///   "code": "OK",
///   "value": {
///       "code": "WalletLoginFailed",
///       "description":"Cannot access the wallet, is the pin correct?"
///    }
/// }
/// ```
NativeString explain_error_code(NativeString error_code);

///
/// Create and initialize, or open and initialize a session (via active agent, to a smart card)
///
/// # Parameters
/// * `handle`: Agent handle
/// * `atr_callback`: Optional seed for default did creation.
/// * `apdu_callback`: A callback that provides a CommandAPDU and expects a ResponseAPDU from the client
///
/// # Example Output
/// ```
/// {
///   "code": "OK",
/// }
/// ```
/// or in case of error
/// ```
/// {
///   "code": "InvalidState",
///   "value": "Akis session already occupied"
/// }
/// ```
/// # Remarks
/// If the session cannot be created, please check out the error code.
///
NativeString init_akis_session(NativeHandle agent_handle,
                               ByteArrayOutCallback atr_callback,
                               ByteArrayInByteArrayOutCallback apdu_callback);

NativeString destroy_akis_session(NativeHandle agent_handle);

NativeString update_balance(NativeHandle agent_handle, NativeString balance);

NativeString read_balance(NativeHandle agent_handle);

#ifdef __cplusplus
}  // extern "C"
#endif  // __cplusplus
